/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/*
    Authorzy:
    Pawe (Googie) Salawa
    boogie@myslenice.one.pl

    Micha (NKG) Niedwiecki
    nkg@poczta.onet.pl

    ChangeLog:

        [v1.19]
            * NKG joins to XMMS module project :) ...with nice start:
            * Titles in CP1250 are converted to ISO8859-2, so Kadu will
              display them properly, either if XMMS does not :) (NKG)
            * Konfiguration of signatures, which will be removed from titles (NKG)
            * Small bugfix (XMMS tab haven't been removed from configuration,
              after module unload) (NKG)

        [v1.18]
            * Added remote control of XMMS from Kadu. In Chat window
              press Left_Win_Key with one of: left, right, up, down,
              enter, backspace.
            * Added OSD advanced configuration.

        [v1.17]
            * NKG's patch applied for Kadu CVS from 02.09.2004.

        [v1.16]
            * Upgraded OSD: now detects if XMMS has just started playing some
              music even if it's same song.
            * Upgraded OSD: added custom colors, icon and font.
            * Fixed enabling statuses.

        [v1.15]
            * Added pseudo-OSD based on Kadu hints.

        [v1.14]
            * New style of getting linker flags (xmms-config --libs).

        [v1.13]
            * Yet another updates for Kadu 0.4.0-cvs (28.07.2004).

        [v1.12]
            * Adrians patch applied acorrding to Kadu 0.4.0-cvs.

        [v1.11]
            * NKGs patch which fixes removeing XMMS tab from configuretion
              dialog and 8bit chars in songs titles. Thanks NKG! :)
            * Fixed requirement of runned XMMS while disabling XMMS statuses.
            * Now XMMS statuses sets empty title status, when XMMS isn't playing.

        [v1.10]
            * Oops, changelog missed ;>

        [v1.9]
            * xmms-config is searched while configuring.

        [v1.8]
            * Added %d tago to descriptions. It's replaced by
              any description you set manualy.
            * Removed warning while compilation.

        [v1.7]
            * Now XMMS tab in Kadu configuration dialog
              is removed after unload module.

        [v1.6]
            * Provides 'descriptions' property and makes
              conflicts with other module which provides this.

        [v1.5]
            * Updated for Kadu 0.3.8-cvs (04.04.2004)

        [v1.4]
            * XMMS menu item changes while XMMS statuses are enabled/disabled.

        [v1.3]
            * Updated for Kadu 0.3.8-cvs (21 II 2004).
            * Fixed for "friends only" statuses.

        [v1.2]
            * Fixed localizations in config dialog.

        [v1.1]
            * Bugfix: Kadu was crashing while getting new message while
              running Kadu with this module.
            * Now Module can be loaded without XMMS runned, but can't be
              used until it isn't runned.
            * Bugfix: Translations wasn't applied to module. Now they are.

        [v1.0]
            * First public release.
*/

#include <xmms/xmmsctrl.h>
#include <stdio.h>

#include "debug.h"
#include "kadu.h"
#include "message_box.h"
#include "userlist.h"
#include "config_dialog.h"
#include "status.h"
#include "../notify/notify.h"

#include "xmms.h"

#define MODULE_XMMS_VERSION 1.19

const char CODE_CP1250[]    = {0xb9,0x9c,0x9f,0xa5,0x8c,0x8f};
const char CODE_ISO8859_2[] = {0xb1,0xb6,0xbc,0xa1,0xa6,0xac};
const char DEFAULT_SIGNATURES[] = "! WWW.POLSKIE-MP3.TK ! \n! www.polskie-mp3.tk ! ";

Xmms *xmms_session;

extern "C" int xmms_init()
{
    xmms_session = new Xmms();

    return 0;
}

extern "C" void xmms_close()
{
    delete xmms_session;
}

// Implementation of Xmms class

Xmms::Xmms()
{
    kdebugf();

    // Disable XMMS statuses by default
    statuses = FALSE;

    // XMMS menus in chats
    menu = new QPopupMenu();
    popups[0] = menu->insertItem(tr("Put formated string"), this, SLOT(putSongTitle(int)));
    popups[1] = menu->insertItem(tr("Put song title"), this, SLOT(putSongTitle(int)));
    popups[2] = menu->insertItem(tr("Put song file name"), this, SLOT(putSongTitle(int)));

    // Title checking timer
    timer = new QTimer();
    connect(timer, SIGNAL(timeout()), this, SLOT(checkTitle()));
    timer->start(1000);

    // XMMS statuses item in Kadu menu
    popups[3] = kadu->mainMenu()->insertItem(tr("Enable XMMS statuses"), this, SLOT(toggleStatuses(int)), 0, -1, 0);

    // Monitor of creating chats
    connect(chat_manager, SIGNAL(chatCreated(const UinsList&)), this, SLOT(chatCreated(const UinsList&)));

    // Checking for %0 in description
    connect(kadu, SIGNAL(changingStatus(UserStatus&, bool&)), this, SLOT(checkStatusDesc(UserStatus&, bool&)));

    // Configuration
    ConfigDialog::addTab("XMMS");
    ConfigDialog::addLineEdit("XMMS", "XMMS", QT_TRANSLATE_NOOP("@default", "XMMS message format for chat windows"), "chatString", "XMMS: %t [%c / %l]",
        QT_TRANSLATE_NOOP("@default", "%t - song title, %f - file name, %l - song length (MM:SS),<br>%c - current song position (MM:SS), %p - percents of played song"));
    ConfigDialog::addLineEdit("XMMS", "XMMS", QT_TRANSLATE_NOOP("@default", "XMMS status format"), "statusString", "-=[%t]=-",
        QT_TRANSLATE_NOOP("@default", "%t - song title, %f - file name, %l - song length (MM:SS),<br>%c - current song position (MM:SS), %p - percents of played song"));

    ConfigDialog::addVGroupBox("XMMS", "XMMS", QT_TRANSLATE_NOOP("@default", "Pseudo-OSD"));
    ConfigDialog::addCheckBox("XMMS", "Pseudo-OSD", QT_TRANSLATE_NOOP("@default", "Enable pseudo-OSD"), "osd", TRUE,
        QT_TRANSLATE_NOOP("@default", "Enables pseudo-OSD done by Kadu hints. Shows currently played title in hint."));
    ConfigDialog::addSelectFont("XMMS", "Pseudo-OSD", QT_TRANSLATE_NOOP("@default", "OSD font"),
        "OsdFont", "helvetica");
    ConfigDialog::addColorButton("XMMS", "Pseudo-OSD", QT_TRANSLATE_NOOP("@default", "OSD font color"), "OsdFgColor",
        QColor("yellow"));
    ConfigDialog::addColorButton("XMMS", "Pseudo-OSD", QT_TRANSLATE_NOOP("@default", "OSD background color"), "OsdBgColor",
        QColor("#000099"));
    ConfigDialog::addSpinBox("XMMS", "Pseudo-OSD", QT_TRANSLATE_NOOP("@default", "OSD timeout"), "OsdTimeout",
        1, 20, 1, 3, QT_TRANSLATE_NOOP("@default", "Time after which OSD disappears (in seconds)"));

    ConfigDialog::addVGroupBox("XMMS", "XMMS", QT_TRANSLATE_NOOP("@default","Cut signatures"));
    ConfigDialog::addCheckBox("XMMS", "Cut signatures", QT_TRANSLATE_NOOP("@default","Enable cut signatures"), "signature", TRUE, "Czy wyci dziada?");
    ConfigDialog::addTextEdit("XMMS", "Cut signatures", QT_TRANSLATE_NOOP("@default","Signatures:"), "signatures",DEFAULT_SIGNATURES);


    // Initial values of some object variables
    descriptionToUse = "";
    descriptionToSet = "";
    winKeyPressed = FALSE;
}

Xmms::~Xmms()
{
    kdebugf();

    ConfigDialog::removeControl("XMMS", "Signatures:");
    ConfigDialog::removeControl("XMMS", "Enable cut signatures");
    ConfigDialog::removeControl("XMMS", "Cut signatures");

    // Removing configuration
    ConfigDialog::removeControl("XMMS", "OSD timeout");
    ConfigDialog::removeControl("XMMS", "OSD font color");
    ConfigDialog::removeControl("XMMS", "OSD background color");
    ConfigDialog::removeControl("XMMS", "OSD font");
    ConfigDialog::removeControl("XMMS", "Enable pseudo-OSD");
    ConfigDialog::removeControl("XMMS", "Pseudo-OSD");

    ConfigDialog::removeControl("XMMS", "XMMS message format for chat windows");
    ConfigDialog::removeControl("XMMS", "XMMS status format");
    ConfigDialog::removeTab("XMMS");

    // Stop timer for checking titles
    timer->stop();

    // Disconnect all slots
    disconnect(timer, SIGNAL(timeout()), this, SLOT(checkTitle()));
    disconnect(chat_manager, SIGNAL(chatCreated(const UinsList&)), this, SLOT(chatCreated(const UinsList&)));
    disconnect(kadu, SIGNAL(changingStatus(UserStatus&, bool&)), this, SLOT(checkStatusDesc(UserStatus&, bool&)));

    // Delete own objects
    delete menu;
    delete timer;

    // Remove Kadu menu item (statuses)
    kadu->mainMenu()->removeItem(popups[3]);
}

void Xmms::chatCreated(const UinsList& ul)
{
    kdebugf();
    Chat* chat = chat_manager->findChatByUins(ul);

    QPushButton* xmms = new QPushButton(chat->buttontray);
    xmms->setPixmap(icons_manager.loadIcon("Unmute"));
    xmms->setPopup(menu);
    xmms->show();
    QToolTip::add(xmms, "XMMS");

    // Key bindings
    connect(chat->edit, SIGNAL(keyPressed(QKeyEvent*, CustomInput*)), this, SLOT(chatKeyPressed(QKeyEvent*, CustomInput*)));
    connect(chat->edit, SIGNAL(keyReleased(QKeyEvent*, CustomInput*)), this, SLOT(chatKeyReleased(QKeyEvent*, CustomInput*)));
}

void Xmms::chatKeyPressed(QKeyEvent* e, CustomInput* k)
{
    kdebugf();
    if (e->key() == Qt::Key_Super_L)
        winKeyPressed = TRUE;
    else if (!winKeyPressed || !xmms_remote_is_running(0))
        return;

    switch (e->key())
    {
        case Qt::Key_Left:
            prevTrack();
            break;
        case Qt::Key_Right:
            nextTrack();
            break;
        case Qt::Key_Up:
            incrVolume();
            break;
        case Qt::Key_Down:
            decrVolume();
            break;
        case Qt::Key_Return:
        case Qt::Key_Enter:
            if (isPlaying())
                pause();
            else
                play();
            break;
        case Qt::Key_Backspace:
            stop();
            break;
    }
}

void Xmms::chatKeyReleased(QKeyEvent* e, CustomInput* k)
{
    if (e->key() == Qt::Key_Super_L)
        winKeyPressed = FALSE;
}

void Xmms::putSongTitle(int ident)
{
    kdebugf();
    if (!xmms_remote_is_running(0))
    {
        MessageBox::msg(tr("XMMS isn't running!"));
        return;
    }
    Chat* chat = getCurrentChat();
    QString title;
    int id;
    for ( int i = 0; i < 3; i++ )
    {
        if (popups[i] == ident)
        {
            id = i;
            break;
        }
    }

    switch (id)
    {
        case 0:
            title = parse(config_file.readEntry("XMMS", "chatString"));
            break;
        case 1:
            title = getTitle();
            break;
        case 2:
            title = getFile();
            break;
    }

    int x, y;

    HtmlDocument doc, doc2, doc3;
    chat->edit->getCursorPosition(&y, &x);
    chat->edit->insertAt(title, y, x);
    doc.parseHtml(chat->edit->text());
    for ( int i = 0; i < doc.countElements(); i++ )
    {
        if (i == 7)
        {
            doc2.parseHtml(doc.elementText(i));
            for ( int j = 0; j < doc2.countElements(); j++ )
            {
                if (doc2.isTagElement(j))
                    doc3.addTag(doc2.elementText(j));
                else
                    doc3.addText(doc2.elementText(j));
            }
        } else {
            if (doc.isTagElement(i))
                doc3.addTag(doc.elementText(i));
            else
                doc3.addText(doc.elementText(i));
        }
    }
    chat->edit->setText(doc3.generateHtml());
    chat->edit->moveCursor(QTextEdit::MoveEnd, FALSE);
}

QString Xmms::parse(QString str)
{
    kdebugf();
    uint sl = str.length();
    QString r;

    for ( uint i = 0; i < sl; i++ )
    {
        while (str[i] != '%' && i < sl)
        {
            r += str[i];
            i++;
        }

        if (str[i] == '%')
        {
            i++;
            switch(str[i].latin1())
            {
                case 't':
                {
                    r += getTitle();
                    break;
                }
                case 'f':
                {
                    r += getFile();
                    break;
                }
                case 'l':
                {
                    char ms[10];
                    int lgt = getLength() / 1000, m, s;
                    m = lgt / 60;
                    s = lgt % 60;
                    if (s < 10)
                        sprintf(ms, "%d:0%d", m, s);
                    else
                        sprintf(ms, "%d:%d", m, s);

                    r += ms;
                    break;
                }
                case 'c':
                {
                    char ms[10];
                    int lgt = getCurrentPos() / 1000, m, s;
                    m = lgt / 60;
                    s = lgt % 60;
                    if (s < 10)
                        sprintf(ms, "%d:0%d", m, s);
                    else
                        sprintf(ms, "%d:%d", m, s);

                    r += ms;
                    break;
                }
                case 'p':
                {
                    char tmp[10];
                    int perc = getCurrentPos() / getLength();
                    sprintf(tmp, "%d%%", perc);
                    r += tmp;
                    break;
                }
                case 'd':
                {
                    r += descriptionToUse;
                    break;
                }
                default:
                    r += str[i];
            }
        }
    }
    return r;
}

Chat* Xmms::getCurrentChat()
{
    kdebugf();
    ChatList cs = chat_manager->chats();
    uint i;
    for ( i = 0; i < cs.count(); i++ )
    {
        if (cs[i]->isActiveWindow())
        {
            break;
        }
    }
    if (i == cs.count())
        return 0;

    return cs[i];
}

QString Xmms::getTitle()
{
    kdebugf();
    if (xmms_remote_is_playing(0))
    {
        QString title(convert(xmms_remote_get_playlist_title(0, pos())));
        if (config_file.readBoolEntry("XMMS","signature",TRUE))
        {
            QStringList sigList(QStringList::split('\n',config_file.readEntry("XMMS","signatures",DEFAULT_SIGNATURES)));
            for (unsigned int i = 0; i < sigList.count(); i++)
            {
                title.remove(sigList[i]);
            }
        }
        return title;
    }
    else
        return "";
}

QString Xmms::getFile()
{
    kdebugf();
    return xmms_remote_get_playlist_file(0, pos());
}

int Xmms::getLength()
{
    kdebugf();
    return xmms_remote_get_playlist_time(0, pos());
}

int Xmms::getCurrentPos()
{
    kdebugf();
    return xmms_remote_get_output_time(0);
}

int Xmms::pos()
{
    return xmms_remote_get_playlist_pos(0);
}

void Xmms::toggleStatuses(int id)
{
    kdebugf();
    if (statuses)
    {
        kadu->mainMenu()->changeItem(popups[3], tr("Enable XMMS statuses"));
        statuses = FALSE;
    }
    else
    {
        if (!xmms_remote_is_running(0))
        {
            MessageBox::msg(tr("XMMS isn't running!"));
            return;
        }
        kadu->mainMenu()->changeItem(popups[3], tr("Disable XMMS statuses"));
        statuses = TRUE;
    }
}

void Xmms::checkTitle()
{
    QString title = getTitle();
    int pos = getCurrentPos();
    if (config_file.readBoolEntry("XMMS", "osd", TRUE) && pos < 1000 && pos > 0)
        putTitleHint(title);

    if (!gadu->status().isOffline() && statuses)
    {
        if (title != currentTitle || !gadu->status().hasDescription())
        {
            currentTitle = title;
            gadu->status().setDescription(parse(config_file.readEntry("XMMS", "statusString")));
        }
    }
}

void Xmms::checkStatusDesc(UserStatus& status, bool &stop)
{
    QString str = config_file.readEntry("XMMS", "statusString");
    if (str.find("%d") > -1 && status.description() != "" && descriptionToSet != status.description())
    {
        stop = TRUE;
        descriptionToUse = status.description();
        QString newDesc = parse(str);
        gadu->status().setDescription(newDesc);
        descriptionToSet = newDesc;
        gadu->status().setStatus(status);
    }
}

void Xmms::putTitleHint(QString title)
{
    kdebugf();
    QColor *fg = new QColor("yellow"), *bg = new QColor("#000099");
    QFont *font = new QFont("helvetica", 14, QFont::Bold);
    QMap<QString, QVariant> parms;
    parms["Pixmap"] = icons_manager.loadIcon("Unmute");
    parms["Foreground color"] = config_file.readColorEntry("XMMS", "OsdFgColor", fg);
    parms["Background color"] = config_file.readColorEntry("XMMS", "OsdBgColor", bg);
    parms["Font"] = config_file.readFontEntry("XMMS", "OsdFont", font);
    parms["Timeout"] = config_file.readNumEntry("XMMS", "OsdTimeout", 3);
    parms["ShowSource"] = bool(FALSE);

    notify->emitMessage("XMMS", "Hints", title, &parms);

    delete fg;
    delete bg;
    delete font;
}

void Xmms::nextTrack()
{
    kdebugf();
    xmms_remote_playlist_next(0);
}

void Xmms::prevTrack()
{
    kdebugf();
    xmms_remote_playlist_prev(0);
}

void Xmms::play()
{
    kdebugf();
    xmms_remote_play(0);
}

void Xmms::stop()
{
    kdebugf();
    xmms_remote_stop(0);
}

void Xmms::pause()
{
    kdebugf();
    xmms_remote_pause(0);
}

void Xmms::setVolume(int vol)
{
    kdebugf();
    xmms_remote_set_main_volume(0, vol);
}

void Xmms::incrVolume()
{
    kdebugf();
    int vol = xmms_remote_get_main_volume(0);
    vol += 2;
    if (vol > 100)
        vol = 100;

    setVolume(vol);
}

void Xmms::decrVolume()
{
    kdebugf();
    int vol = xmms_remote_get_main_volume(0);
    vol -= 2;
    if (vol < 0)
        vol = 0;

    setVolume(vol);
}

bool Xmms::isPlaying()
{
    kdebugf();
    return xmms_remote_is_playing(0);
}

QString Xmms::convert(QString str)
{
//    QString str(src);
    for (unsigned int i = 0; i < sizeof(CODE_CP1250); i++)
    {
        str.replace(CODE_CP1250[i],CODE_ISO8859_2[i]);
    }
    return QString::fromLocal8Bit(str,-1);
}
